# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 15:04:15 2016

@author: Tobias Jachowski
"""
import numpy as np
#import os
import struct

'''
def get_labview_bin_data_shape(filename):
    print('Getting data shape from:')
    print('  \'%s\'' % filename)
    filesize = os.path.getsize(filename) / 8  # size in bytes
    chunk_shape = np.fromfile(filename, count=2, dtype='>i4')
    # size of buffer, which is read in every loop iteration
    buffer_size = chunk_shape[0]
    # number of channels read in
    number_channels = chunk_shape[1]
    # chunk size
    chunk_size = 1 + buffer_size * number_channels
    # chunk_shape of single chunks is not changed during a measurement with the
    # ForceClamp -> same chunk_size for all chunks
    # number of chunks in file
    number_chunks = int(filesize/(chunk_size))
    number_samples = buffer_size * number_chunks

    shape_file = (number_chunks, chunk_size)
    shape_record = (number_samples, number_channels)

    return shape_file, shape_record


def read_labview_bin_data(filename, dtype='>d'):
    # big-endian double-precision floating-point number
    shape_file, shape_record = get_labview_bin_data_shape(filename)

    # 1. read in data
    # 2. reshape to two dimensions: chunks and bytes(/samples)
    # 3. remove byte corresponding to the chunk shape information
    # 4. reshape data to: number of samples and number of channels
    count = shape_file[0] * shape_file[1]
    # workaround, if file filesize % chunk_size is not an integer
    # it leaves out additional numbers

    print("Reading data from:\n  '%s'" % filename)
    data = np.fromfile(
        filename,
        dtype=dtype,
        count=count).reshape(shape_file)[:, 1:].reshape(shape_record)
    # data = np.loadtxt(filename, comments='%', delimiter=' ')

    # Workaround, if last column is 0.0
    if np.all(data[:, -1] == 0.0):
        # kick out last column
        data = data[:, :-1]

    # data = np.memmap(filename,
    #                  dtype=dtype,
    #                  mode='r', shape=(l, k))[:,1:].reshape([l*i, j])
    #
    # fname_o = ''
    # convert_bin_data(filename, fname_o, k, 1)
    # data = np.memmap(fname_o, dtype=dtype, mode='r', shape=(l*i, j))
    # data = np.fromfile(fname_o, dtype=dtype).reshape([l*i, j])
    return data
'''


def chunk_info(filename):
    """
    Read all chunk arrays with the information (number of rows and columns) of
    the data, following each chunk.
    """
    with open(filename, "rb") as f:
        pos = 0
        while True:
            byte = f.read(8)
            pos += 1
            if byte:
                chunk_shape = struct.unpack('>2i', byte)
                if chunk_shape[0] * chunk_shape[1] > 0:
                    # go to byte offset relative to current strem position
                    f.seek(8 * chunk_shape[0] * chunk_shape[1], 1)
                    yield(pos, *chunk_shape)
                    pos += chunk_shape[0] * chunk_shape[1]
            else:
                break


def chunk_data(filename, chunks, dtype='>d'):
    with open(filename, "rb") as f:
        for chunk in chunks:
            # go to byte offset where chunk starts
            f.seek(chunk[0] * 8)
            # read the data
            data_bin = f.read(chunk[1] * chunk[2] * 8)
            data = np.fromstring(data_bin, dtype=dtype)
            yield data


def read_labview_bin_data(filename, dtype='>d', start_row_idx=0,
                          number_of_rows=-1):
    """
    Parameters
    ----------
    filename : str
        Path of the binary file to read
    dtype : str
        Type of double
    start_row_idx : int
        Index of the first datapoint (of all traces) to read
    number_or_rows : int
        Number of datapoints (of all traces) to read. Defaults to number of
        datapoints of the binary file - `start_row_idx`
    """
    stop_row_idx = start_row_idx + number_of_rows  # index of row to stop read
    rows_to_read = 0  # rows to read after getting information of chunks
    columns = None  # number of columns (i.e. traces) in binary file
    chunks = []  # chunks with the index information of data to be read
    chunk_row_start = 0  # running index of first row of chunk
    chunk_row_stop = 0  # running stop index of last row of chunk

    if number_of_rows == 0:
        return np.empty((0, 0))

    print('Getting chunk info from:')
    print('  \'%s\'' % filename)
    for chunk in chunk_info(filename):
        # Set the running indices to the new position (row)
        chunk_row_start = chunk_row_stop
        chunk_row_stop += chunk[1]
        # Check if information about number of columns changed from one to the
        # next chunk
        if columns == chunk[2] or columns is None:
            columns = chunk[2]
        else:
            print("Number of columns in chunks of file differ from each other!")
        # Check if chunk is (partly) contained within the requested data index
        if (chunk_row_stop > start_row_idx
                and (number_of_rows < 0 or chunk_row_start < stop_row_idx)):
            # Check read position and number of rows of first chunk
            if len(chunks) is 0:
                shift = start_row_idx - chunk_row_start
                chunk = (chunk[0] + shift * chunk[2],
                         chunk[1] - shift,
                         chunk[2])
            chunks.append(chunk)
            rows_to_read += chunk[1]
        # Check stop position and number of rows of last chunk
        if (number_of_rows > 0 and chunk_row_stop >= stop_row_idx):
            shift = chunk_row_stop - stop_row_idx
            chunk = (chunk[0], chunk[1] - shift, chunk[2])
            chunks[-1] = chunk
            rows_to_read -= shift
            break

    if rows_to_read == 0:
        return np.empty((0, 0))

    print('Reading data chunks from:')
    print('  \'%s\'' % filename)
    data = np.empty(rows_to_read * columns)
    i = 0
    for _data in chunk_data(filename, chunks, dtype=dtype):
        length = len(_data)
        data[i:i + length] = _data
        i += length

    data = data.reshape(rows_to_read, columns)

    # Workaround, if last column is 0.0
    if np.all(data[:, -1] == 0.0):
        # kick out last column
        data = data[:, :-1]

    return data


def __OLD_read_labview_bin_data(filename, dtype='>d'):
    rows = 0
    columns = None
    chunks = []

    print('Getting chunk info from:')
    print('  \'%s\'' % filename)
    for chunk in chunk_info(filename):
        c = chunk[2]
        if columns == c or columns is None:
            columns = c
        else:
            print("Number of columns in chunks of file differ from each other!")
        rows += chunk[1]
        chunks.append(chunk)

    print('Reading data chunks from:')
    print('  \'%s\'' % filename)
    data = np.empty(rows * columns)
    i = 0
    for _data in chunk_data(filename, chunks, dtype=dtype):
        length = len(_data)
        data[i:i + length] = _data
        i += length

    data = data.reshape(rows, columns)

    # Workaround, if last column is 0.0
    if np.all(data[:, -1] == 0.0):
        # kick out last column
        data = data[:, :-1]

    return data


def write_labview_bin_data(filename, array, dtype='d', endianness='>'):

    i = struct.pack('>i', array.shape[0])
    j = struct.pack('>i', array.shape[1])
    fmt = endianness + dtype * array.size
    data = struct.pack(fmt, *array.flatten())

    with open(filename, 'wb') as f:
        f.write(i)
        f.write(j)
        f.write(data)

'''
def convert_bin_data(fname_i, fname_o, chunk_size, chunk_shape_size=1):
    """
    Remove chunk size information of labview binary files.
    """
    fi = open(fname_i, 'rb')
    fo = open(fname_o, 'wb')

    while True:
        buf = fi.read(chunk_size)
        if len(buf) == 0:
            break
        fo.write(buf[chunk_shape_size:])

        fi.close()
        fo.close()
'''


# -*- coding: utf-8 -*-
"""
Created on Wed Mar  9 17:09:09 2016

@author: Tobias Jachowski
"""
import collections
import numpy as np
try:
    import pandas as pd
    __pd__ = True
except ImportError:
    __pd__ = False
from abc import ABCMeta, abstractmethod, abstractproperty

from .. import helpers as hp
from .. import traces as tc
from ..graph import GraphMember
from ..evaluate import signal as sn


class Region(GraphMember, metaclass=ABCMeta):
    """
    A Region provides common methods, to define and reference regions on given
    data, and methods to access the data in the defined regions.
    A Region of the data to refer to can be choosen by setter methods
    (start/stop or tmin/tmax). The region can be defined as a timespan
    (seconds) or an indexspan (indexes of a numpy array).
    """

    def __init__(self, caching=False, **kwargs):
        """
        Parameters
        ----------
        caching : bool
        **kwargs

        Attributes
        ----------
        start
        stop
        tmin
        tmax
        timespan
        samplingrate
        datapoints
        seconds
        indexspan
        indexvector
        timevector
        num_traces
        caching
        traces
        data
        """
        super().__init__(**kwargs)
        self.caching = caching

    @abstractproperty
    def start(self):
        pass

    @abstractproperty
    def stop(self):
        pass

    @start.setter
    def start(self, start):
        pass

    @stop.setter
    def stop(self, stop):
        pass

    @property
    def tmin(self):
        return self.start / self.samplingrate

    @property
    def tmax(self):
        return (self.stop - 1) / self.samplingrate

    @tmin.setter
    def tmin(self, tmin):
        self.start = int(np.around(tmin * self.samplingrate))

    @tmax.setter
    def tmax(self, tmax):
        self.stop = int(np.around(tmax * self.samplingrate)) + 1

    @property
    def timespan(self):
        return (self.tmin, self.tmax)

    @timespan.setter
    def timespan(self, timespan):
        self.tmin, self.tmax = timespan

    def set_timespan(self, tmin, tmax):
        self.tmin = tmin
        self.tmax = tmax

    @abstractproperty
    def samplingrate(self):
        # points per second
        pass

    @property
    def resolution(self):
        return self.samplingrate

    @property
    def datapoints(self):
        return self.stop - self.start

    @property
    def seconds(self):
        return self.datapoints / self.samplingrate

    @property
    def indexspan(self):
        # return slice(self.start, self.stop)
        return slice(0, self.datapoints, 1)

    @property
    def indexvector(self):
        # return np.arange(self.start, self.stop)
        return np.arange(self.datapoints)

    @property
    def timevector(self):
        return self.indexvector / self.samplingrate

    @abstractproperty
    def num_traces(self):
        return self.data.shape[0]

    @property
    def caching(self):
        return self._caching

    @caching.setter
    def caching(self, caching):
        """
        If memory consumption is no concern, use caching to speed up data
        retrieval.
        """
        self._caching = caching
        if not caching:
            # delete cache to free memory, ZODB volatile
            self._v_data_cached = None

    def get_data(self, traces=None, samples=None, moving_filter='mean',
                 window=1, decimate=1, copy=True, pandas=False):
        """
        Returns data of this region.

        Parameters
        ----------
        traces : str or list of str, optional
        samples : int, list of int, or slice, optional
            Samples start/stop has to be in between 0 and self.datapoints. If
            start, stop or any other index is negative it will be converted to
            a positive index, referring to self.datapoints.
        moving_filter : str, optional
            The moving filter to apply to the data. Can be one of 'mean' or
            'median'. Only applied, if `window` > 1.
        window : int, optional
            The window size of the moving filter.
        decimate : int, optional
            Decimate the data by given value. If `samples` is of type slice and
            `samples.step` is explicitly set (i.e. not None), the step value
            takes precedence over `decimate`.
        copy: bool, optional
            Only if `self.caching` is True, `copy` has the effect of either
            returning a copy of the cached data (`copy`=True) or returning a
            reference to the cached data array.
            If `self.caching` is False, return the uncached data. Then, `copy`
            has no effect at all.
        pandas : bool, optional
            Return a pandas.DataFrame with proper indexes and colunmns.

        Returns
        -------
        2D numpy.ndarray or pandas.DataFrame
            If self.caching is True, return cached data.
        """
        # Set the traces and the samples to proper default values and/or
        # convert (normalize) them to proper formats
        traces_idx = self.traces_to_idx(traces)
        samples = self.samples_idx(samples, decimate)

        if window <= 1:
            # Return unfiltered data
            data = self._get_data(samples, traces_idx, copy)
        else:
            # Moving filter should be applied, filter first, then decimate.
            # Get data, disregarding decimating factor, but considering traces
            if isinstance(samples, slice):
                samples_start = samples.start
                samples_stop = samples.stop
            else:  # samples is numpy.ndarray
                samples_start = samples.min()
                samples_stop = samples.max()
            # Get requested and preceding and succeeding data, to avoid
            # boundary issues of the filter
            filter_start = max(0, samples_start - window)
            filter_stop = min(samples_stop + window, self.datapoints)
            filter_samples = slice(filter_start, filter_stop, 1)
            data = self._get_data(filter_samples, traces_idx, copy=False)

            # Filter the data
            data = hp.moving_filter(data, window, moving_filter=moving_filter)

            # Return requested samples, considering decimating factor
            if isinstance(samples, slice):
                start = samples_start - filter_start
                stop = samples_stop - filter_start
                samples = slice(start, stop, samples.step)
            else:  # samples is numpy.ndarray
                samples = samples - filter_start
            data = data[samples]

        if pandas and __pd__:
            return pd.DataFrame(data, index=self.timevector[samples],
                                columns=self.idx_to_traces(traces_idx))

        return data

    def _get_data(self, samples, traces_idx, copy=True):
        """
        Returns the data for the given samples and traces_idx.
        """
        # Retrieve cached data
        if self.caching:
            # If the cache is outdated or not yet created, do so
            self.update_cache()

            # Check, if traces and/or samples are instances of numpy.ndarray
            traces_is_ndarray = isinstance(traces_idx, np.ndarray)
            samples_is_ndarray = isinstance(samples, np.ndarray)

            # If at least one of the indices is an ndarray, advanced indexing
            # is triggered.
            if traces_is_ndarray or samples_is_ndarray:
                # If both, traces and samples are an ndarray, indexing two
                # dimensions in a single step would result in an error in
                # numpy. Therefore, separate the indexing into two steps.
                if traces_is_ndarray and samples_is_ndarray:
                    # First, select the samples.
                    data_ = self._v_data_cached[samples]

                    # Second, select the traces.
                    return data_[:, traces_idx]

                # Advanced indexing automatically returns a copy of the data.
                # There is no need to extra copy via ndarray.copy(), which
                # would cost some extra execution time.
                copy = False

            if copy:
                # Per default return a copy of the data cache to protect it
                # from unwanted in place modifications
                return self._v_data_cached[samples, traces_idx].copy()
            else:
                # Only, if explicitly asked for "copy=False", return a
                # direct reference to the cached data
                return self._v_data_cached[samples, traces_idx]

        # No caching enabled, return uncached data
        return self._get_data_uncached(samples, traces_idx, copy)

    def update_cache(self, force=False):
        """
        Create and update cached data.

        Parameters
        ----------
        force : bool
            Recalculate, even if self._v_data_cached is up to date.
        """
        # Check for whether an update of the cache is needed or not.
        if self.caching and (not hasattr(self, '_v_data_cached')
                             or self._v_data_cached is None
                             or force):
            # An update is needed, calculate data for self.indexspan and
            # all traces_idx and store it in the cache. ZODB volatile.
            self._v_data_cached = self._get_data_uncached(self.indexspan,
                                                          self.traces_to_idx(),
                                                          copy=True)

    @abstractmethod
    def _get_data_uncached(self, samples, traces_idx, copy=True):
        pass

    def member_changed(self, ancestor=True, calledfromself=False,
                       leave_cache=False, **kwargs):
        # If self triggered a change, delete cache and trigger an update of the
        # cache. `leave_cache` can prevent deleting of the cache.
        if calledfromself and not leave_cache:
            self._v_data_cached = None  # ZODB volatile
        # If an ancestor triggered a change, delete cache and trigger an update
        # of the cache. A triggered change of descendants is ignored.
        if not calledfromself and ancestor:
            self._v_data_cached = None  # ZODB volatile

        # Call method of superclass `GraphMember`
        super().member_changed(ancestor=ancestor,
                               calledfromself=calledfromself, **kwargs)

    def traces_available(self, traces=None):
        traces_idx = self.traces_to_idx(traces)
        return self.idx_to_traces(traces_idx)

    def traces_to_idx(self, traces=None):
        """
        return index/slice of trace/s
        if trace is None, default to all traces
        """
        # Default to all traces
        if traces is None:
            traces_idx = slice(0, self.num_traces, 1)
        # If traces is a slice, correct for negative and None values
        elif isinstance(traces, slice):
            # TODO: if step is negative, indices() evaluates stop to
            # -1, but it should be -1 -self.num_traces!
            start, stop, step = traces.indices(self.num_traces)
            traces_idx = slice(start, stop, step)
        else:
            traces_idx = []
            for trace in tc.normalize(traces):
                if trace in self.traces:
                    traces_idx.append(self.traces.index(trace))
                elif isinstance(trace, int) and trace < self.num_traces:
                    traces_idx.append(trace)

        return hp.slicify(traces_idx, length=self.num_traces)

    def idx_to_traces(self, traces_idx=None):
        if isinstance(traces_idx, slice):
            traces_idx = hp.listify(traces_idx)
        if traces_idx is None:
            traces = self.traces
        elif isinstance(traces_idx, collections.Iterable):
            traces = [self.traces[trace_idx]
                      for trace_idx in traces_idx
                      if trace_idx < self.num_traces]
        return traces

    def samples_idx(self, samples=None, decimate=1):
        """
        Returns a slice or an np.array.

        Parameters
        ----------
        samples : int or list of int or slice, optional
            If samples is None, return the slice spanning the whole Region:
            slice(0, `self.datapoints`, `decimate`).
            If `samples` is an instance of int, slice, list or np.array,
            negative indices (i_neg) are converted into positive ones (i_pos):
            i_pos = i_neg + `self.datapoints`
        decimate : int, optional
            The decimate value of the returned slice, or the value the index
            array is decimated with.
            If `samples` is of type slice and `samples.step` is not explicitly
            set (i.e. None), `decimate` takes precedence over the step value.
        """
        if samples is None:
            samples = slice(0, self.datapoints, decimate or 1)
        elif isinstance(samples, int):
            if samples < 0:
                # correct for negative indices in original samples
                # referring to self.datapoints
                samples += self.datapoints
            samples = slice(samples, samples + 1, decimate or 1)
        elif isinstance(samples, slice):
            # Correct samples for negative indices and None
            start, stop, step = samples.indices(self.datapoints)

            # If samples.step is not explicitly set (i.e. None), decimate takes
            # precedence of the step information (1 per default)
            step = samples.step or decimate or 1

            samples = slice(start, stop, step)
        else:
            # samples either a list or an np.array
            # convert list to array and apply decimate (1 per default)
            if isinstance(samples, list):
                samples = np.array(samples)[::decimate or 1]
            # correct for negative indices in original samples
            # referring to self.datapoints
            samples[samples < 0] += self.datapoints

        return hp.slicify(samples)

    def _excited(self, traces=None, index=True):
        """
        For backward compatibility in pyotc.

        Parameters
        ----------
        traces : list of str
            Two traces to be compared. Defaults to ['positionX', 'positionY'].

        Returns
        -------
        trace : str
            The excited trace.
        """
        traces = traces or ['positionX', 'positionY']
        data = self.get_data(traces=traces, copy=False)
        ex = sn.get_excited_signal(data)
        if index:
            return ex
        return traces[ex]

    @abstractproperty
    def traces(self):
        return

    @property
    def data(self):
        # return all data
        return self.get_data()

    def __getattr__(self, name):
        """
        Allow attributes to be used as trace selections for get_data
        """
        if name in self.traces or name in tc:
            # name is directly known by Region or
            # name is probably an alias, a shorthand notation, or a combination
            return self.get_data(traces=name)
        else:
            raise AttributeError(name)



# -*- coding: utf-8 -*-
"""
Created on Thu May 22 19:49:29 2014

@author: Tobias Jachowski
"""
import collections
import io
import os
import numpy as np
try:
    import pandas as pd
    __pd__ = True
except ImportError:
    __pd__ = False
import sys
from scipy.ndimage import convolve1d
from scipy.ndimage import median_filter


if sys.platform == "linux" or sys.platform == "linux2":
    platform = 'linux'
elif sys.platform == "darwin":
    platform = 'darwin'
elif sys.platform == "win32":
    platform = 'win32'


def get_png_image(figure):
    figure.canvas.draw()
    buf = io.BytesIO()
    figure.savefig(buf, format='png')
    png = buf.getvalue()
    buf.close()
    return png


def slicify(index, length=-1):
    """
    Takes an index as an instance of a list, a tuple, an np.ndarray, an int, or
    anything else. In the case of a list, a tuple, or a np.ndarray return a
    slice if possible, else return an np.ndarray. In the case of an int, return
    slice(int, int+1). In the case of anything else, return anything else.
    """
    # Convert list and tuple in an np.ndarray
    if isinstance(index, list) or isinstance(index, tuple):
        index = np.array(index)
    if isinstance(index, np.ndarray):
        # No elements in the array
        if len(index) == 0:
            return slice(0, 0, 1)
        # One element in the array
        if len(index) == 1:
            return slice(index[0], index[-1] + 1, 1)
        # More elements in the array
        if len(index) > 1:
            diff = np.diff(index)
            # Accept only strictly monotonically indices
            if diff[0] != 0 and np.all(diff[:-1] == diff[1:]):
                step = diff[0]
                start = index[0]
                stop = index[-1] + np.sign(step)
                # If the last element is at index 0 and the step is negative,
                # try to figure out the negative stop index. This is due
                # to the fact that a slice object behaves different for
                # numpy.ndarray than for the function `range()`.
                if stop < 0 and length < 0:
                    stop = step * len(index) - 1
                if stop < 0 and length >= 0:
                    stop = - length - 1
                return slice(start, stop, step)
    elif isinstance(index, int):
        return slice(index, index + 1, 1)

    return index


def listify(trace, length=-1):
    """
    Takes a trace as an instance of int, slice, or list and returns a list
    """
    if isinstance(trace, int) or isinstance(trace, str):
        return [trace]
    if isinstance(trace, slice):
        if trace.stop < 0 and trace.step < 0 and length < 0:
            length = abs(trace.stop) - 1
        elif trace.stop < 0 and trace.step > 0 and length < 0:
            length = abs(trace.stop) - 1
        elif trace.stop >= 0 and trace.step < 0 and length < 0:
            length = trace.start + trace.stop + 1
        elif trace.stop >= 0 and trace.step > 0 and length < 0:
            length = trace.stop
        else:  # length >= 0:
            length = length
        start, stop, step = trace.indices(length)
        return list(range(start, stop, step))
    if isinstance(trace, np.ndarray):
        return trace.tolist()
    return trace


def missing_elements(list1, list2):
    idx_existent = np.in1d(list1, list2)
    idx_missing = np.logical_not(idx_existent)
    missing_elements = listify(np.array(list1)[idx_missing])
    return missing_elements


def overlap_index(list1, list2):
    return np.nonzero(np.in1d(list1, list2))[0]


def skip_index(list1, list2):
    return np.nonzero(np.logical_not(np.in1d(list1, list2)))[0]


def moving_filter(data, window, moving_filter='mean', mode='reflect', cval=0.0,
                  origin=0):
    """
    Apply a moving filter to data.

    Parameters
    ----------
    data : numpy.ndarray
        The data to be filterd.
    window : int
        The window size of the moving filter.
    moving_filter : str, optional
        The filter to be used for the moving filter. Can be one of 'mean' or
        'median'.
    mode : str, optional
        mode       |   Ext   |         Data           |   Ext
        -----------+---------+------------------------+---------
        'mirror'   | 4  3  2 | 1  2  3  4  5  6  7  8 | 7  6  5
        'reflect'  | 3  2  1 | 1  2  3  4  5  6  7  8 | 8  7  6
        'nearest'  | 1  1  1 | 1  2  3  4  5  6  7  8 | 8  8  8
        'constant' | 0  0  0 | 1  2  3  4  5  6  7  8 | 0  0  0
        'wrap'     | 6  7  8 | 1  2  3  4  5  6  7  8 | 1  2  3
        See 'scipy.ndimage.convolve1d' or 'scipy.ndimage.median_filter'
    cval : float, optional
        See 'scipy.ndimage.convolve1d' or 'scipy.ndimage.median_filter'
    origin : int, optional
        See 'scipy.ndimage.convolve1d' or 'scipy.ndimage.median_filter'
    """
    mode = mode or 'reflect'
    cval = cval or 0.0
    origin = origin or 0
    if moving_filter == 'mean' or moving_filter == 'average':
        return movingmean(data, window, mode=mode, cval=cval, origin=origin)
    else:  # if moving == 'median'
        return movingmedian(data, window, mode=mode, cval=cval, origin=origin)


def movingmean(data, window, mode='reflect', cval=0.0, origin=0):
    weights = np.repeat(1.0, window)/window
    # sma = np.zeros((data.shape[0] - window + 1, data.shape[1]))
    sma = convolve1d(data, weights, axis=0, mode=mode, cval=cval,
                     origin=origin)
    return sma


def movingmedian(data, window, mode='reflect', cval=0.0, origin=0):
    if data.ndim == 1:
        size = window
    else:
        size = (window, 1)
    smm = median_filter(data, size=size, mode=mode, cval=cval, origin=origin)
    return smm


def moving_mean(data, window):
    """
    Calculate a filtered signal by using a moving mean. The first datapoint is
    the mean of the first `window` datapoints and the last datapoint is the mean
    of the last `window` datapoints of the original data. This function does not
    handle the lost edges of the data, i.e. the filtered data is shortened by
    `window` datapoints.

    This function is faster than the function `movingmean()`.

    Parameters
    ----------
    data : 1D numpy.ndarray of type float
        Data to calculate the rolling mean from.
    window : int
        Length of the window to calculate the rolling mean with.

    Returns
    -------
    1D numpy.ndarray of type float
        The data filtered with a rolling mean.
    """
    cumsum = np.cumsum(np.insert(data, 0, 0))
    return (cumsum[window:] - cumsum[:-window]) / window


def moving_mean_pandas(data, window):
    """
    Calculate a filtered signal by using a moving mean.

    Parameters
    ----------
    data : 1D numpy.ndarray of type float
        Data to calculate the rolling mean from.
    window : int
        Length of the window to calculate the rolling mean with.

    Returns
    -------
    1D numpy.ndarray of type float
        The data filtered with a rolling mean.
    """
    data = pd.Series(data)
    r = data.rolling(window=window)
    return r.mean()[window - 1:].get_values()


def calculate_means(data, samples=None, stds=False):
    """
    Calculate the means of `data`. If `samples_idx` is given, calculate the
    means of all samples contained in `samples_idx`.
    Additionally, the standard deviations can be returned, if stds is set to
    True.

    Parameters
    ----------
    data : numpy.ndarray
    samples : int, slice, or Iterable of int, slice, or index arrays
        Samples to calcualte the means from.
    stds : bool, optional
        Return also the standard deviations for the calculated means.

    Returns
    -------
    numpy.ndarray
        The calculated means
    (numpy.ndarray, numpy.ndarray)
        If `stds` is True, the calculated means and standard deviations.
    """
    if samples is None:
        samples = [slice(0, len(data))]
    if not isinstance(samples, collections.Iterable):
        samples = [samples]

    means = np.array([data[s].mean(axis=0) for s in samples], ndmin=2)

    if stds:
        stds = np.array([data[s].std(axis=0, ddof=1) for s in samples], ndmin=2)
        return means, stds

    return means


def bin_means(data, number_of_bins=None, datapoints_per_bin=None,
              sortcolumn=0):
    """
    Calculate binned means.

    Parameters
    ----------
    data : 2D numpy.ndarray of type float
    number_of_bins : int, optional
        Number of bins the datapoints should be averaged with. Defaults to
        sqrt(len(data)), if datapoints_per_bin is None.
    datapoints_per_bin : int, optional
        Average number of datapoints to be averaged in one bin.
    sortcolumn : int, optional
        Column of `data` that acts as sorting index upon binning for the rest
        of the data. Defaults to the first column (`data[:,0])`.

    Returns
    -------
    2D numpy.ndarray of type float
        The averaged bin values.
    1D numpy.ndarray of type float
        The center of the bins
    float
        The width of the bins.
    """
    # Calculate number of bins
    if number_of_bins is None:
        if datapoints_per_bin is None:
            number_of_bins = round(np.sqrt(len(data)))
        else:
            number_of_bins = max(1, round(len(data) / datapoints_per_bin))
    number_of_bins = int(number_of_bins)

    # Create the bins based on data[:, sortcolumn]
    minimum = data[:, sortcolumn].min()
    maximum = data[:, sortcolumn].max()
    bin_edges = np.linspace(minimum, maximum, number_of_bins + 1)
    bin_width = bin_edges[1] - bin_edges[0]
    bin_centers = bin_edges[0:-1] + bin_width / 2

    # Get the indices of the bins to which each value in input array belongs.
    bin_idx = np.digitize(data[:, sortcolumn], bin_edges)

    # Calculate the means of the data in the bins
    bin_means = np.array([data[bin_idx == i].mean(axis=0)
                          for i in range(1, len(bin_edges))])
    bin_Ns = np.array([np.sum(bin_idx == i)
                       for i in range(1, len(bin_edges))])
    bin_stds = np.array([data[bin_idx == i].std(axis=0, ddof=1)
                         for i in range(1, len(bin_edges))])

    return bin_centers, bin_means, bin_stds, bin_Ns, bin_width


def file_and_dir(filename=None, directory=None):
    filename = filename or ""
    fdir = os.path.dirname(filename)
    ffile = os.path.basename(filename)

    ddir = directory or "."

    if (ffile == "" or ffile == "." or ffile == ".."):
        directory = os.path.join(ddir, filename, "")
        absdir = os.path.realpath(directory)
        return None, absdir, None

    directory = os.path.join(ddir, fdir, "")
    absdir = os.path.realpath(directory)
    absfile = os.path.join(absdir, ffile)

    return ffile, absdir, absfile


def files(directory, prefix=None, suffix=None, extension=None, sort_key=None):
    """
    Get filenames of a directory in the order sorted to their filename or a
    given key function.

    Parameters
    ----------
    directory : str
        The directory the files are located in.
    prefix : str
        Get only the files beginning with `prefix`.
    suffix : str
        Get only the files ending with `suffix`.
    extension : str, optional
        The extension of the files that should be returned. Default is
        '.txt'.
    sort_key : function
        Function to be applied to every filename found, before sorting.
    """
    prefix = prefix or ''
    suffix = suffix or ''
    extension = extension or ''
    files = [file_and_dir(filename=name, directory=directory)[2]
             for name in os.listdir(directory)
             if os.path.isfile(os.path.join(directory, name))
             and name.startswith(prefix)
             and name.endswith(''.join((suffix, extension)))]
    files.sort(key=sort_key)
    return files



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 14:29:39 2016

@author: Tobias Jachowski
"""
import numpy as np
import operator
from collections import namedtuple
from lmfit import minimize, Parameters, fit_report
from scipy import constants


k_B = constants.value('Boltzmann constant')
ForceExtension = namedtuple('ForceExtension', ['extension', 'force'])


def worm_like_chain(x, L_0, L_p=50e-9, T=298):
    """
    A worm-like chain model.

    Marko, J.F.; Eric D. Siggia. "Stretching DNA". Macromolecules. 1995. 28:
    8759–8770. doi:10.1021/ma00130a008

    Parameters
    ----------
    x : float
        Extension (m)
    L_0 : float
        Contour length (m)
    L_p : float
        Persistence length (m)
    T : float
        Temperature (K)

    Returns
    -------
    1D numpy.ndarray of type float
        Force (N).
    """
    F = k_B * T / L_p * (1 / (4 * (1 - x / L_0)**2) - 1/4 + x / L_0)

    return F


def worm_like_chain_1p(x, L_0, L_p=50e-9, T=298):
    """
    An improved worm-like chain model.

    Petrosyan, R. "Improved approximations for some polymer extension
    models". Rehol Acta. 2016. doi:10.1007/s00397-016-0977-9

    Parameters
    ----------
    x : float
        Extension (m)
    L_0 : float
        Contour length (m)
    L_p : float
        Persistence length (m)
    T : float
        Temperature (K)

    Returns
    -------
    1D numpy.ndarray of type float
        Force (N).
    """
    F = k_B * T / L_p * (1 / (4 * (1 - x / L_0)**2) - 1/4 + x / L_0
                         - 0.8 * (x / L_0)**2.15)

    return F


# TODO: Check units
def twistable_wlc(F, L_0, x=None, L_p=43.3e-9, K_0=1246e-12, S=1500e-12,
                  C=440e-30, T=298.2):
    """
    The twistable worm-like chain model.

    Gross, P.; Laurens, N.; Oddershede, L. B.; Bockelmann, U.; Peterman, E. J.
    & Wuite, G. J. "Quantifying how DNA stretches, melts and changes twist
    under tension". Nature Physics, Nature Research, 2011, 7, 731-736

    Paramaters
    ----------
    F : float
        force (N)
    L_0 : float
        contour length (m)
    x : float
        extension (m)
    L_p : float
        persistence length (m)
    K_0 : float
        elastic modulus (N)
    S : float
        stretch modulus (N)
    C : float
        twist rigidity (Nm²)
    """
    def g(F, fitting=False, use_lambda_fit=False, use_unwinding_fit=False):
        """
        g(F) describes the twist-stretch coupling how DNA complies to tension.

        The original equation is as follows:
        g(F) = (S * C - C * F * (x/L_0 - 1 + 1/2
                * (k*T/(F*L_p))**(1/2))**(-1))**(1/2)

        The equation can be simplified:
            - below Fc of 30 pN: as a constant (- 100 pN nm)
            - above Fc to linear order: g0 + g1 * F

        g0 was determied to be 590 pN nm (or 560 pN nm with lambda DNA)
        g1 was determined to be 18 nm
        """
        if fitting:
            return (S * C - C * F * (x/L_0 - 1 + 1/2
                    * (k_B*T/(F*L_p))**(1/2))**(-1))**(1/2)
        if F <= 30e-12:  # N
            return - 100e-21  # Nm
        else:
            g0 = - 590e-21  # Nm
            if use_lambda_fit:
                g0 = - 560e-21  # Nm
            if use_unwinding_fit:
                g0 = - 637e-21  # Nm
            return g0 + 17e-9 * F

    x = L_0 * (1 - 1/2 * (k_B * T / (F * L_p))**(1/2)
               + C / (-g(F)**2 + S * C) * F)

    return x


def force_extension(bps, pitch=0.338e-9, L_p=43.3e-9, T=298.2, min_ext=0.5,
                    max_ext=0.978, samples=1000):
    """
    Calculate force extension data for a DNA, assuming a worm-like chain model.

    Parameters
    ----------
    bps : int
        number of base-pairs
    pitch : float
        length of one base-pair (m)
        Defaults to 0.338e-9 m (Saenger 1988)
    L_p : float
        persistence length (m)
    T : float
        temperature (K)
    min_ext : float
        minimum extension, normalized to contour length
        0.5-0.978 entspricht ~520nm-1057nm (0.118...49.2pN)
    max_ext : float
        maximum extension, normalized to contour length
    samples : int
        number of samples to generate

    Returns
    -------
    ForceExtension
        The namedtuple consists of extension and force in (m)
    """
    L_0 = bps * pitch  # m contour length
    x = np.linspace(min_ext * L_0, max_ext * L_0, samples)
    F = worm_like_chain(x, L_0, L_p, T)

    # K_0=1246e-12  # N elastic modulus
    # x = (x + f / K_0) * L_0

    return ForceExtension(extension=x, force=F)


def twistable_force_extension(bps, pitch=0.338e-9, L_p=43.3e-9, K_0=1246e-12,
                              S=1500e-12, C=440e-30, T=298.2, min_f=10e-12,
                              max_f=60e-12, samples=1000):
    F = np.linspace(min_f, max_f, samples)
    L_0 = bps * pitch
    x = twistable_wlc(F, L_0, L_p, K_0, S, C, T)

    return ForceExtension(extension=x, force=F)


def crop_x_y(x, y=None, min_x=None, max_x=None, min_y=None, max_y=None,
             include_bounds=True):
    """
    Crop pairs of variates according to their minimum and maximum values.

    Parameters
    ----------
    x : 1D numpy.ndarray of type float
        The x values.
    y : 1D numpy.ndarray of type float
        The y values.
    min_x : float
        The minimum value of `x`.
    max_x : float
        The maximum value of `x`.
    min_y : float
        The minimum value of `y`.
    max_y : float
        The maximum value of `y`.
    include_bounds : bool
        Whether to include or exlude min/max values in the output arrays.

    Returns
    -------
    tuple of 2 1D numpy.ndarray of type float
        The cropped values (x, y).
    """
    # Reduce calculation time, if no min/max values are given
    if min_x is None and max_x is None and min_y is None and max_y is None:
        if y is None:
            return x
        else:
            return x, y
    if include_bounds:
        ol = operator.le
        og = operator.ge
    else:
        ol = operator.lt
        og = operator.gt
    max_x = max_x or float('inf')
    min_x = min_x or float('-inf')
    i_x = ol(x, max_x)
    i_x = np.logical_and(i_x, og(x, min_x))
    if y is None:
        return x[i_x]
    max_y = max_y or float('inf')
    min_y = min_y or float('-inf')
    i_y = ol(y, max_y)
    i_y = np.logical_and(i_y, og(y, min_y))
    i = np.logical_and(i_x, i_y)
    return x[i], y[i]


def residual(params, model_func, x, data, min_x_param=None, max_x_param=None,
             eps=None):
    """
    Calculate the residuals of a model and given data.

    Parameters
    ----------
    params : lmfit.Parameters
    model_func : function
        Function with the header model_func(x, **params). It should return
        a 1D numpy.ndarray of type float.
    x : 1D numpy.ndarray of type float
    data : 1D numpy.ndarray of type float
    min_x_param : str
        Crop `x` (and `data`) with the function `crop_x_y()` (paramater
        `min_x`) according to the parameter with the key `min_x_param` in
        `params`,  before calculating the model.
    max_x_param : str
        Crop `x` (and `data`) with the function `crop_x_y()` (paramater
        `max_x`) according to the parameter with the key `max_x_param` in
        `params`,  before calculating the model.
    eps : float
    """
    # Crop the X data according to a fit parameter
    if min_x_param is not None or max_x_param is not None:
        min_x = params.get(min_x_param, None)
        max_x = params.get(max_x_param, None)
        x, data = crop_x_y(x, data, min_x=min_x, max_x=max_x,
                           include_bounds=False)

    # Calculate data according to the model function
    model = model_func(x, **params)

    # Calculate the residuals of the model and the given data
    if eps is None:
        return model - data
    return (model - data) / eps


def get_DNA_fit_params(bps, pitch=None, L_p=None, T=None, vary=None, **kwargs):
    """
    Create a default `lmfit.Parameters` instance, which can be used to fit DNA
    force extension data with lmfit.

    The default parameters calculated are the contour length (L_0), the
    persistence length (L_p) and the temperature (T). The contour length is
    calculated from the number of basepairs (bps) and the pitch (pitch) of the
    DNA.

    Parameters
    ----------
    bps : float
        Number of basepairs of the DNA to be fitted. The parameter `bps` is
        used to calculate the initial value (start value) of the contour length
        'L_0' (i.e. `bps`*`pitch`). See also parameter `params`.
    pitch : float (optional)
        Pitch in nm (defaults to 0.338e-9 nm). The parameter `pitch` is
        used to calculate the initial value (start value) of the contour length
        'L_0' (i.e. `bps`*`pitch`). See also parameter `params`.
    L_p : float (optional)
        Persistence length in m (defaults to 0.338e-9 m).
    T : float
        Temperature in K (defaults to 298 K).
    vary : dict (optional)
        A dictionary setting the Parameters, which should be varied during the
        fitting process (defaults to {'L_0': True, 'L_p': True, 'T': False}).
        As an alternative to the key 'L_0', 'bps' can be used.
    **kwargs
        Additional parameters for compatibility reasons that are neglected.

    Returns
    -------
    lmfit.Parameters
        Lmfit Parameters instance containing the Parameter objects. The default
        is a Parameters instance, containing the Parameter objects 'L_0' (i.e.
        `bps`*`pitch`), `L_p`, and `T`, initialized with the corresponding
        function parameters.
    """
    pitch = pitch or 0.338e-9
    L_p = L_p or 50e-9
    T = T or 298
    # Calculate the contour length of the DNA
    L_0 = bps * pitch  # m contour length

    vary = vary or {}
    vary_L_0_default = vary.get('bps', True)

    params = Parameters()
    params.add('L_0', value=L_0, vary=vary.get('L_0', vary_L_0_default))
    params.add('L_p', value=L_p, vary=vary.get('L_p', True))
    params.add('T', value=T, vary=vary.get('T', False))

    return params


def fit_force_extension(e, f, bps, model_func=None, params=None, min_e=None,
                        max_e=None, max_f=None, max_e_dyn_L0=True,
                        verbose=False, return_model_func=False, **kwargs):
    """
    Fit a model function, e.g. a worm-like chain model, to force extension data
    of DNA.

    Parameters
    ----------
    e : 1D numpy.ndarray of type float
        The extension (m).
    f : 1D numpy.ndarray of type float
        The force (N).
    bps : float
        The number of basepairs of the DNA. If you do not know the number of
        basepairs, try an upper estimate to fit the data. The number of
        basepairs is used to calculate the start value for the contour length
        L_0 for the fitting procedure. See also the function
        `get_DNA_fit_params.
    model_func : func
        Set model function, that should have the header model_func(e, **params)
        and return f. Defauts to the function `worm_like_chain`.
    params : lmfit.Parameters
        The parameters to be fitted (defaults to the output of
        `get_DNA_fit_params(bps, **kwargs)`).
    min_e : float
        Minimum extension in m (defaults to float(-inf)) to be used to fit the
        model.
    max_e : float
        Maximum extension in m (defaults to float(+inf)) to be used to fit the
        data. See also parameter `max_e_dyn_L0`.
    max_f : float
        Maximum force in N (defaults to 15e-12) to be used to fit the data.
    max_e_dyn_L0 : bool
        Set `max_e` dynamically to the contour length 'L_0' for every fitting
        evaluation. If L0 is greater than the parameter `max_e`, the value of
        `max_e` will be used, instead.
    verbose : bool
        Be verbose about the fit result.
    return_model_func : bool
        Return the used model_func additionally to the fit result.
    **kwargs
        Arguments passed to the parameter function `get_DNA_fit_params`.

    Returns
    -------
    lmfit.minimizer.MinimizerResult
        If `return_model_func` is False.
    (lmfit.minimizer.MinimizerResult, model_func)
        If `return_model_func` is True
    """
    # Choose the model function and initialize the fit parameters
    model_func = model_func or worm_like_chain
    params = params or get_DNA_fit_params(bps, **kwargs)

    # Crop the data. Choose boundaries to avoid nan values and a max force of
    # 15 pN, up to where the wlc model is valid.
    min_x = min_e
    max_x = max_e
    max_y = max_f or 15e-12
    e, f = crop_x_y(e, f, min_x=min_x, max_x=max_x, max_y=max_y,
                    include_bounds=False)

    # Choose the parameters for the function residual, which calculates the
    # difference of the model function to the given data
    residual_args = model_func, e, f
    residual_kwargs = {}
    if max_e_dyn_L0:
        # Crop e and f data variates to contour length 'L_0', i.e. up to where
        # the wlc model is valid
        residual_kwargs['max_x_param'] = 'L_0'

    # Do the fitting:
    #   minimize -> residual(params, residual_args) -> model_func(e, params)
    out = minimize(residual, params, args=residual_args, kws=residual_kwargs)

    if verbose:
        print(fit_report(out))
        print('[[DNA related info]]')
        print('    Number of base-pairs: {:.0f}'.format(
                            np.round(out.params['L_0'] / params['L_0'] * bps)))

    if return_model_func:
        return out, model_func
    else:
        return out



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 13:36:39 2016

@author: Tobias Jachowski
"""
import matplotlib.pyplot as plt
import numpy as np
from ipywidgets import interact, IntSlider

from . import signal as sn
from . import dna
from . import evaluate
from .. import gui
from .. import helpers as hp
from .evaluate import Evaluator
from .signalfeature import CycleSectioner


class Tether(Evaluator):
    """
    Calculates tether related parameters, like sections of stressing or
    releasing the bead, bead-center surface distance, extension of the tether,
    or force acting on the tether.

    To distinguish between the excited axes ('x', 'y'), directions
    ('left', 'right'), and different cycles ('stress', 'release'), a
    CycleSectioner is used.
    """
    def __init__(self, region=None, calibration=None, resolution=1000,
                 filter_time=-1.0, resolution_sf=None, filter_time_sf=None,
                 **kwargs):
        """
        This __init__() constructor extends the superclass (Evaluator)
        constructor by setting the traces_sf to default to 'positionXY'.

        For extensive documentation of the Parameters and additional
        attributes, see `pyoti.evaluate.evaluate.Evaluator`.

        Parameters
        ----------
        region : pyoti.region.region.Region
        calibration : pyoti.calibration.calibration.Calibration
        resolution : float
        filter_time : float
        resolution_sf : float
        filter_time_sf : float
        **kwargs
            Extra arguments, e.g. traces_sf, in case the positionXY signals are
            called differently.

        Attributes
        ----------
        displacementXYZ : 2D numpy.ndarray
        forceXYZ : 2D numpy.ndarray
        force
        distanceXYZ
        distance
        extension
        angle
        rightstress
        leftstress
        leftrelease
        rightrelease
        stress
        release
        rightstresses
        leftstresses
        leftreleases
        rightreleases
        stresses
        releases
        """

        traces_sf = kwargs.pop('traces_sf', 'positionXY')

        # resolution for (almost) all properties, e.g. displacement, force, ...
        # resolution 1.000 Hz, samplingrate 40.000 Hz -> 40 points
        super().__init__(region=region, calibration=calibration,
                         resolution=resolution, filter_time=filter_time,
                         sf_class=CycleSectioner, traces_sf=traces_sf,
                         resolution_sf=resolution_sf,
                         filter_time_sf=filter_time_sf)

        self.fe_figure = None

    def _sections(self, axis=None, direction=None, cycle=None,
                  concat_direction=False, concat_cycle=False, info=False,
                  **kwargs):
        """
        Calculate start/stop indices (segments) of the requested axis,
        direction, and cycle (sections).

        Parameters
        ----------
        axis : str or list of str, optional
            'x', 'y', or ['x', 'y'] (default)
        direction : str or list of str, optional
            'left', 'right', or ['left', 'right'] (default)
        cycle : str or list of str, optional
            'stress', 'release', or ['stress', 'release'] (default)
        concat_direction : bool, optional
            Concatenate individual left/right sections into one section. Only
            evaluated, if `concat_cycle` is True.
        concat_cycle : bool, optional
            Concatenate individual stress/release sections into one new section
            with the start beeing the start of stress and the stop beeing the
            stop of release.
        info : bool, optional
            Additionally to the sections return a 2D numpy.ndarray, containing
            the axis, direction, and the cycle for every individual section.
            Axis can be either 'x' or 'y'. Direction can be 'left', 'right' or
            'leftright'. Cycle can be 'stress', 'release', or 'stressrelease'.

        Returns
        -------
        2D np.ndarray or tuple of two 2D np.ndarray
            If info is False start/stop values in rows.
            If info is True start/stop values in rows and an 2D np.ndarray with
            infos (see parameter `info`).
        """
        if axis is None:
            axis = ['x', 'y']
        if direction is None:
            direction = ['left', 'right']
        else:
            # A direction was chosen, prevent concat
            concat_direction = False
        if isinstance(direction, str):
            direction = [direction]
        if cycle is None:
            cycle = ['stress', 'release']
        else:
            # A cycle was chosen, prevent concat
            concat_cycle = False
        if isinstance(cycle, str):
            cycle = [cycle]

        sections = np.empty((0, 2), dtype=int)
        infos = np.empty((0, 3), dtype=str)

        # Get all requested segments
        # for ax, axi in zip(axis, range(len(axis))):
        for ax in axis:
            # separate release/stress and separate left/right -> x/y,
            # right/left, stress/release
            if not concat_cycle:
                for sec, dc in zip([d + c for d in direction for c in cycle],
                                   [[d, c] for d in direction for c in cycle]):
                    for segment in self._sf.sections[ax]:
                        sections = np.r_[sections, segment[sec]]
                        if len(segment[sec]) > 0:
                            seg_info = np.array([[ax, dc[0], dc[1]]]
                                                * len(segment[sec]))
                            infos = np.r_[infos, seg_info]
            # concat release/stress and separate left/right -> x/y, right/left
            if concat_cycle and not concat_direction:
                # for d, di in zip(direction, range(len(direction))):
                for d in direction:
                    for segment in self._sf.sections[ax]:
                        sections = np.r_[sections, segment[d]]
                        if len(segment[d]) > 0:
                            seg_info = np.array([[ax, d, 'stressrelease']]
                                                * len(segment[d]))
                            infos = np.r_[infos, seg_info]
            # concat release/stress and concat left/right -> x/y
            if concat_cycle and concat_direction:
                sections = np.r_[sections, self._sf.excited[ax]]
                if len(self._sf.excited[ax]) > 0:
                    seg_info = np.array([[ax, 'leftright', 'stressrelease']]
                                        * len(self._sf.excited[ax]))
                    infos = np.r_[infos, seg_info]

        if info:
            return sections, infos
        else:
            return sections

    def _extrema(self, axis=None, extremum=None):
        if axis is None:
            axis = ['x', 'y']
        if extremum is None:
            extremum = ['minima', 'maxima']
        elif isinstance(extremum, str):
            extremum = [extremum]

        extrema = np.empty(0, dtype=int)

        # Get all extrema and sort
        for ax in axis:
            for section in self._sf.sections[ax]:
                for ext_type in extremum:
                    idx = section[ext_type]
                    idx = self.undecimate_and_limit(idx)
                    extrema = np.r_[extrema, idx]
        extrema.sort()
        return extrema

    def stress_release_pairs(self, axis=None, direction=None, i=None,
                             slices=True, decimate=None, info=False, **kwargs):
        """
        Calculate start/stop indices (as segments or slices) of stress/release
        cycle pairs of the requested axis and direction.

        Parameters
        ----------
        axis : str or list of str, optional
            'x', 'y', or ['x', 'y'] (default)
        direction : str or list of str, optional
            'left', 'right', or ['left', 'right'] (default)
        i : int, optional
            The index of the stress release pair to be selected and returned
            from all stress release pairs calculated. Depending on `axis` and
            `direction` the total availabe number of stress release pairs may
            vary.
        slices : bool, optional
            Set to False to return segments instead of slices. Default is True.
        decimate : int, optional
            Used to set the step attribute of the returned slices. Only
            evaluated, if `slices` is True.
        info : bool, optional
        **kwargs : dict, optional
            Only used for compatibility purposes, to be able to call method
            with parameters not defined in method definition.

        Returns
        -------
        stresses : 1D np.ndarray of slices or segments
        releases : 1D np.ndarray of slices or segments
        stress_infos : 1D np.ndarray of 1D np.ndarrays of type str
            The str arrays have the form of [axis, direction, cycle]. Only
            returned, if parameter `info` is True.
        release_infos : 1D np.ndarray of 1D np.ndarrays of type str
            The str arrays have the form of [axis, direction, cycle]. Only
            returned, if parameter `info` is True.
        """
        if axis is None:
            axis = ['x', 'y']
        if direction is None:
            direction = ['left', 'right']
        if isinstance(direction, str):
            direction = [direction]
        stress_segments, _stress_infos = self.sections(axis=axis,
                                                       direction=direction,
                                                       cycle='stress',
                                                       slices=False,
                                                       info=True)
        release_segments, _release_infos = self.sections(axis=axis,
                                                         direction=direction,
                                                         cycle='release',
                                                         slices=False,
                                                         info=True)
        extrema = self._extrema(axis=axis)
        stresses = np.empty((0, 2), dtype=int)
        releases = np.empty((0, 2), dtype=int)
        stress_infos = np.empty((0, 3), dtype=str)
        release_infos = np.empty((0, 3), dtype=str)

        # Group all stress/release cycle pairs according to the extrema
        # A stress release pair corresponds to one extremum, only if the stop
        # of the stress and the start of the release segment equal the
        # extremum. However, either one of the stress or release segment can be
        # missing.
        for ext in extrema:
            # Find stress whose stop is equal to extremum
            stress_idx = stress_segments[:, 1] == ext
            stress = stress_segments[stress_idx]
            stress_info = _stress_infos[stress_idx]
            # Find release whose start is equal to extremum
            release_idx = release_segments[:, 0] == ext
            release = release_segments[release_idx]
            release_info = _release_infos[release_idx]
            # Create stress slice, if stress is empty, but release is valid
            if stress.size == 0 and release.size > 0:
                stress = np.array([[ext, ext]])
            # Create release slice, if release is empty, but stress is valid
            if release.size == 0 and stress.size > 0:
                release = np.array([[ext, ext]])
            if stress.size > 0 and release.size > 0:
                # p_p = np.r_[stress, release]  # concatenate stress/release
                # cycle pair
                # concatenate with other cycles
                stresses = np.r_[stresses, stress]
                releases = np.r_[releases, release]
                stress_infos = np.r_[stress_infos, stress_info]
                release_infos = np.r_[release_infos, release_info]

        # Convert segments into slices
        if slices:
            stresses = sn.idx_segments_to_slices(stresses, decimate=decimate)
            releases = sn.idx_segments_to_slices(releases, decimate=decimate)

        # Get the maximum number of stress release pairs
        stop = len(stresses)
        # Prevent index overflow and allow negative indices
        if i is not None:
            if i < 0:
                i = stop + i
            i = max(0, i)
            i = min(i, stop - 1)
            s = slice(i, i + 1)
        else:
            s = slice(0, stop)

        if info:
            return stresses[s], releases[s], stress_infos[s], release_infos[s]
        else:
            return stresses[s], releases[s]

    def baseline_idx(self, axis=None, strict=False, extrapolate=False):
        """
        Calculate the indices of the baseline, according to the excited axes,
        i.e. the indices, where the excited axis is zero. This method uses
        the detected sections of `self.region`, instead of searching for
        the value 0.0 in the excited axes, as compared to the function
        `pyoti.evaluate.signal.basline_idx()`.

        Parameters
        ----------
        axis : str or list of str, optional
            'x', 'y', or ['x', 'y'] (default)
        strict : bool
            Take only baseline indices of releases's stops or stresses's
            starts, if a stress directly follows a release section.
        extrapolate : bool
            Take baseline indices, even if they are indexed by only one
            stress or release segment, i.e. indices which mark the start or
            the end of a wave.
        """
        # Get stress/release pairs
        if strict:
            # Get only stress/release pairs, i.e. only segment pairs, whose
            # release's stop and stress's start index equal an extremum.
            stresses, releases = self.stress_release_pairs(axis=axis,
                                                           slices=False)
        else:
            # Get all stress/release segments, even those without a
            # corresponding extremum.
            stresses = self.sections(axis=axis, cycle='stress', slices=False)
            releases = self.sections(axis=axis, cycle='release', slices=False)

        # A baseline point lies exactly beetween a release and a following
        # stress section. Therefore, the baseline point idx equals a stress's
        # start and a release's stop.
        stress_base_idx = stresses[:, 0]
        release_base_idx = releases[:, 1]

        # Remove stresses, whose start, and releases, whose stop is equal to
        # an extremum, which means, the excitation as above or below the
        # baseline.
        extrema = self._extrema(axis=axis)
        stress_base_idx = np.setdiff1d(stress_base_idx, extrema,
                                       assume_unique=True)
        release_base_idx = np.setdiff1d(release_base_idx, extrema,
                                        assume_unique=True)

        base_idx = np.r_[stress_base_idx, release_base_idx]
        if strict or not extrapolate:
            # Take only baseline point indices, that lie exactly between a
            # release and a following stress section, i.e. have two equal
            # entries, one from a release's start and one from a stress's stop.
            # I.e. sort out all indices that did not come from a stress/release
            # pair but instead only a single stress or release segment.
            base_idx.sort()
            base_idx = base_idx[np.r_[base_idx[:-1] == base_idx[1:], False]]
        return np.unique(base_idx)

    def _rfigure(self, legend=True, fig=None, ax=None):
        if fig is None and ax is None:
            fig, ax = plt.subplots()
            suptitle = True
        elif fig is None:
            fig = ax.get_figure()
            suptitle = False
        elif ax is None:
            ax = fig.gca()
            suptitle = False

        ax.grid(True)

        line_rstr = None
        line_rrls = None
        line_lstr = None
        line_lrls = None
        line_minima = None
        line_maxima = None
        t = self.timevector
        for axis, trace in zip('xy', ['positionX', 'positionY']):
            s = self.get_data(trace)
            rstr, rrls = self.stress_release_pairs(axis=axis,
                                                   direction='right')
            lstr, lrls = self.stress_release_pairs(axis=axis, direction='left')

            ax.plot(t, s, lw=0.1, ms=2, color='k', alpha=1.0)

            # line_rstr = None
            # line_rrls = None
            # line_lstr = None
            # line_lrls = None
            for rstr, rrls in zip(rstr, rrls):
                line_rstr, = ax.plot(t[rstr], s[rstr], lw=0.4, ms=2, color='m')
                line_rrls, = ax.plot(t[rrls], s[rrls], lw=0.4, ms=2, color='c')
            for lstr, lrls in zip(lstr, lrls):
                line_lstr, = ax.plot(t[lstr], s[lstr], lw=0.4, ms=2, color='g')
                line_lrls, = ax.plot(t[lrls], s[lrls], lw=0.4, ms=2, color='y')

            # line_minima = None
            # line_maxima = None
            for segment in self._sf.sections[axis]:
                minima = self.undecimate_and_limit(segment['minima'])
                maxima = self.undecimate_and_limit(segment['maxima'])
                line_minima, = ax.plot(t[minima], s[minima], '.', ms=5,
                                       color='b')
                line_maxima, = ax.plot(t[maxima], s[maxima], '.', ms=5,
                                       color='r')

        line_excited_x = None
        for x_c in (self.undecimate_and_limit(self._sf.excited['x'])
                    / self.resolution):
            line_excited_x = ax.hlines(0.0, x_c[0], x_c[1], alpha=1,
                                       colors='b', linestyle='solid', lw=1)
            # ax.plot(x_c[0], 0.5, '.k', alpha=1, ms=3)
            # ax.plot(x_c[1], 0.5, '.k', alpha=1, ms=3)
            ax.vlines(x_c[0], -0.01, 0.01, alpha=1, colors='b',
                      linestyle='solid', lw=1)
            ax.vlines(x_c[1], -0.01, 0.01, alpha=1, colors='b',
                      linestyle='solid', lw=1)

        line_excited_y = None
        for y_c in (self.undecimate_and_limit(self._sf.excited['y'])
                    / self.resolution):
            line_excited_y = ax.hlines(0.0, y_c[0], y_c[1], alpha=1,
                                       colors='r', linestyle='solid', lw=1)
            # ax.plot(y_c[0], -0.5, '.k', alpha=1, ms=3)
            # ax.plot(y_c[1], -0.5, '.k', alpha=1, ms=3)
            ax.vlines(y_c[0], -0.01, 0.01, alpha=1, colors='r',
                      linestyle='solid', lw=1)
            ax.vlines(y_c[1], -0.01, 0.01, alpha=1, colors='r',
                      linestyle='solid', lw=1)

        ax.set_xlim((t[0], t[-1]))

        ax.set_ylabel("Signal positionX and Y (um)")
        ax.set_xlabel("Time (s)")
        if suptitle:
            fig.suptitle("Automatically detected excited axis, minima, "
                         "maxima, and sections.")

        if legend:
            if line_minima is not None:
                line_minima.set_label('minima')
            if line_maxima is not None:
                line_maxima.set_label('maxima')
            if line_rstr is not None:
                line_rstr.set_label('rightstress')
            if line_rrls is not None:
                line_rrls.set_label('rightrelease')
            if line_lstr is not None:
                line_lstr.set_label('leftstress')
            if line_lrls is not None:
                line_lrls.set_label('leftrelease')
            if line_excited_x is not None:
                line_excited_x.set_label('excited x')
            if line_excited_y is not None:
                line_excited_y.set_label('excited y')

            ax.legend(loc='upper right')

        return fig

    def force_extension_pair(self, i, axis=None, direction=None, decimate=None,
                             time=False, twoD=False):
        """
        Calculate the force extension pair with index `i`.

        Parameters
        ----------
        i : int
            Index of the force_extension pair to be returned.
        axis : str
            See method `self.stress_release_pairs()`.
        direction : str
            See method `self.stress_release_pairs()`.
        decimate : int
            See method `self.stress_release_pairs()`.

        Returns
        -------
        1D numpy.ndarray of type float
            Extension values of stress cycles in nm.
        1D numpy.ndarray of type float
            Force values of stress cycles in pN.
        1D numpy.ndarray of type str
            (str, str, str), containing the axis, direction, and the cycle.
            Axis can be either 'x' or 'y'. Direction can be 'left', or 'right'.
            Cycle is 'stress'.
        1D numpy.ndarray of type float
            Extension values of release cycles in nm.
        1D numpy.ndarray of type float
            Force values of release cycles in pN.
        1D numpy.ndarray of type str
            (str, str, str), containing the axis, direction, and the cycle.
            Axis can be either 'x' or 'y'. Direction can be 'left', or 'right'.
            Cycle is 'release'.
        """
        fe_pair = self.force_extension_pairs(axis=axis, direction=direction,
                                             i=i, decimate=decimate, time=time,
                                             twoD=twoD)
        return next(fe_pair)

    def force_extension_pairs(self, axis=None, direction=None, i=None,
                              decimate=None, time=False, twoD=False):
        """
        Return a generator for force extension values of stress release pairs.

        Parameters
        ----------
        str_rls_pairs : tuple of 4 numpy.ndarrays, optional
        axis : str, optional
            See method `self.stress_release_pairs()`.
        direction : str, optional
            See method `self.stress_release_pairs()`.
        i : int
            Index of the force extension pair to be yielded.
        decimate : int, optional
            See method `seld.stress_release_pairs()`.

        Yields
        ------
        1D numpy.ndarray of type float
            Extension values of stress cycles in nm.
        1D numpy.ndarray of type float
            Force values of stress cycles in pN.
        1D numpy.ndarray of type str
            (str, str, str), containing the axis, direction, and the cycle.
            Axis can be either 'x' or 'y'. Direction can be 'left', or 'right'.
            Cycle is 'stress'.
        1D numpy.ndarray of type float
            Extension values of release cycles in nm.
        1D numpy.ndarray of type float
            Force values of release cycles in pN.
        1D numpy.ndarray of type str
            (str, str, str), containing the axis, direction, and the cycle.
            Axis can be either 'x' or 'y'. Direction can be 'left', or 'right'.
            Cycle is 'release'.
        """
        str_rls_pairs = self.stress_release_pairs(axis=axis,
                                                  direction=direction,
                                                  i=i,
                                                  decimate=decimate,
                                                  slices=True,
                                                  info=True)
        strs, rlss, stri, rlsi = str_rls_pairs

        # Get the start from the first stress and the stop from the last
        # release cycle
        start = strs[0].start
        stop = rlss[-1].stop
        samples = slice(start, stop)

        # Get extension, force, and stress/release pairs
        e_f = self._force_extension(samples=samples, twoD=twoD) * 1000  # nm,pN
        e = e_f[:, 0]
        f = e_f[:, 1]

        # Get the time
        if time:
            t = self.timevector[samples]

        # Yield all stress/release extension/force data pairs
        for st, rl, sti, rli in zip(strs, rlss, stri, rlsi):
            _st = slice(st.start - start, st.stop - start, st.step)
            _rl = slice(rl.start - start, rl.stop - start, rl.step)
            est = e[_st]
            fst = f[_st]
            erl = e[_rl]
            frl = f[_rl]
            if time:
                tst = t[_st]
                trl = t[_rl]
                yield est, fst, sti, erl, frl, rli, tst, trl
            else:
                yield est, fst, sti, erl, frl, rli

    def init_fe_fig(self, show=True, autolimit=True, plot_params=None):
        """
        Initialize the figure where the force extension pairs should be plotted
        to.

        Parameters
        ----------
        show : bool, optional
            Show the figure in the notebook after it has been initialized.
        autolimit : bool, optional
            Set the limits of the force extension plot according to the min/max
            values of the force and extension.
        plot_params : dict, optional
        """
        xlim = None
        ylim = None
        if self.fe_figure is not None:
            ax = self.fe_figure.axes[0]
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
            plt.close(self.fe_figure)

        # Initialize proper plot parameters
        gui.set_plot_params(plot_params=plot_params)

        # Create static figure
        self.fe_figure = _force_extension_figure()

        if autolimit:
            self.autolimit()
        else:
            ax = self.fe_figure.axes[0]
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)

        if show:
            self.fe_figure.show()

    def autolimit(self, samples=None, e=None, f=None, xlim=None, ylim=None,
                  set_limits=True):
        """
        Determine xlim and ylim values for the force extension pair plots.

        Parameters
        ----------
        samples : int, slice or index array, optional
            Samples to get extension and force from.
        e : 1D numpy.ndarray of floats, optional
            Extension in nm. Takes precedence over extension determined with
            `samples`.
        f : 1D numpy.ndarray of floats, optional
            Force in pN. Takes precedence over force determined with `samples`.
        xlim : (float, float), optional
            Xlimit of force extension axis. Takes precedence over xlim
            determined with `e`.
        ylim : (float, float), optional
            Ylimit of force extension axis. Takes precedence over ylim
            determined with `f`.
        set_limits : bool, optional
            Set xlim and ylim of the force extension plot.

        Returns
        -------
        (float, float)
            The xlim
        (float, float)
            The ylim
        """
        if samples is None \
                and (xlim is None and e is None) \
                or (ylim is None and f is None):
            # Get the start/stop indices of the data to be used to determine
            # the min max values
            sts, rls = self.stress_release_pairs()
            start = sts[0].start
            stop = rls[-1].stop
            samples = slice(start, stop)

        if xlim is None and ylim is None and e is None and f is None:
                e_f = self._force_extension(samples=samples) * 1000  # nm, pN
                e = e_f[:, 0]
                f = e_f[:, 1]
        if xlim is None and e is None:
            e = self._extension(samples=samples) * 1000  # nm
        if ylim is None and f is None:
            f = self._force(samples=samples) * 1000  # pN

        if xlim is None:
            e_min = e.min()
            e_max = e.max()
            e_diff = (e_max - e_min) * 0.02
            xlim = (e_min - e_diff, e_max + e_diff)

        if ylim is None:
            f_min = f.min()
            f_max = f.max()
            f_diff = (f_max - f_min) * 0.02
            ylim = (f_min - f_diff, f_max + f_diff)

        # Set the limits
        if set_limits:
            fig = self.fe_figure
            ax = fig.axes[0]
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            fig.canvas.draw()

        # Return the set limits
        return xlim, ylim

    def update_force_extension_plot(self, fe_pair=None, xlim=None, ylim=None,
                                    title=None, draw=True, bps=None):
        """
        Update the figure with force extension data.

        Parameters
        ----------
        fe_pair : tuple of 6 1D numpy.ndarrays, optional
            As returned by the method `self.force_extension_pair()`.
        xlim : (float, float), optional
            Set xlim of the axis.
        ylim : (float, float), optional
            Set ylim of the axis.
        title : str
            Set the title of the figure.
        draw : bool
            Redraw, after the the data has been updated.
        bps : int
            If lenght of DNA in basepairs is given, a standard stretching
            curve is plotted.
        """
        if self.fe_figure is None:
            self.init_fe_fig(show=False)

        fig = self.fe_figure
        ax = fig.axes[0]

        if xlim is not None:
            ax.set_xlim(xlim)
        if ylim is not None:
            ax.set_ylim(ylim)

        # Get stress/release extension/force data pair
        if fe_pair is not None:
            # Unpack force_extension_pair and set the line data to the unpacked
            # values
            time = False
            if len(fe_pair) == 8:
                time = True

            if time:
                est, fst, sti, erl, frl, rli, tst, trl = fe_pair
            else:
                est, fst, sti, erl, frl, rli = fe_pair
            ax.lines[0].set_data(est, fst)
            ax.lines[1].set_data(erl, frl)
            # Autodetermine the title
            if time:
                title = title or "Force Extension, %s, %s, (%.3f, %.3f) s" \
                                 % (sti[0], sti[1], tst[0], trl[-1])
            else:
                title = title or "Force Extension, %s, %s" % (sti[0], sti[1])

        # Calculate force extension of a dna with a known length and plot it
        if bps:
            x, F = dna.force_extension(bps=bps)
            ax.lines[2].set_data(x*1e9, F*1e12)
        else:
            ax.lines[2].set_data([0], [0])

        if title is not None:
            ax.set_title(title)

        if draw:
            fig.canvas.draw()

    def plot_force_extension(self, i, autolimit=True, xlim=None, ylim=None,
                             bps=None, draw=True, **kwargs):
        """
        Plot the force extension data with index `i` (see method
        `self.force_extension_pairs()`) on self.fe_figure.

        Parameters
        ----------
        i : int
            Index of force extension pair. See method
            `self.force_extension_pair()`.
        autolimit : bool, optional
            Automatically detect xlim and/or ylim from min/max values of force
            and extension. The limits are only detected, if `xlim` and/or
            `ylim` are not supplied.
        xlim : (float, float), optional
            Xlimit of force extension axis.
        ylim : (float, float), optional
            Ylimit of force extension axis.
        bps : int, optional
            Base pairs of standard force extension curve of a dsDNA to be
            plotted along with the experimental data.
        draw : bool, optional
            Update the figure, after data has been plotted.
        **kwargs : dict, optional
            Keyword arguments passed to `self.force_extension_pair()`.
        """
        fe_pair = self.force_extension_pair(i, **kwargs)
        e = np.r_[fe_pair[0], fe_pair[3]]
        f = np.r_[fe_pair[1], fe_pair[4]]
        time = kwargs.get('time', False)

        if autolimit:
            xlim, ylim = self.autolimit(e=e, f=f, xlim=xlim, ylim=ylim,
                                        set_limits=False)
        if time:
            title = "Force Extension %.3i, %s, %s, (%.3f, %.3f) s" \
                    % (i, fe_pair[2][0], fe_pair[2][1], fe_pair[6][0],
                       fe_pair[7][-1])
        else:
            title = "Force Extension %.3i, %s, %s" % (i, fe_pair[2][0],
                                                      fe_pair[2][1])

        self.update_force_extension_plot(fe_pair=fe_pair, xlim=xlim, ylim=ylim,
                                         title=title, draw=draw, bps=bps)

    def plot_force_extensions(self, autolimit=True, xlim=None, ylim=None,
                              bps=None, draw=True):
        """
        Plot force/extension stress/release pairs on an axis of a figure.

        Parameters
        ----------
        autolimit : bool, optional
            Automatically detect xlim and/or ylim from min/max values of force
            and extension. The limits are only detected, if `xlim` and/or
            `ylim` are not supplied.
        xlim : (float, float), optional
            Xlimit of force extension axis.
        ylim : (float, float), optional
            Ylimit of force extension axis.
        bps : int, optional
            Base pairs of standard force extension curve of a dsDNA to be
            plotted along with the experimental data.
        draw : bool, optional
            Update the figure, after data has been plotted.

        Yields
        ------
        matplotlib.figure.Figure
            Yield a reference to the figure, after the next force extension
            force/extension stress/release data pair has been plotted.
        """
        # Get xlim and ylim according to min/max values of the force extension
        # pairs to be plotted
        if autolimit:
            xlim, ylim = self.autolimit(xlim=xlim, ylim=ylim, set_limits=False)

        # Get force extension pairs generator
        fe_pairs = self.force_extension_pairs()

        # Plot all stress/release extension/force data pairs
        for idx, fe_pair in enumerate(fe_pairs):
            title = "Force Extension %.3i, %s, %s" % (idx, fe_pair[2][0],
                                                      fe_pair[2][1])
            self.update_force_extension_plot(fe_pair=fe_pair, xlim=xlim,
                                             ylim=ylim, title=title, draw=draw,
                                             bps=bps)
            yield self.fe_figure

    def show_force_extension_plots(self, **kwargs):
        """
        Display a slider to show the force extensions created by
        `self.plot_force_extension()`.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments passed to `self.plot_force_extension()`.

        Returns
        -------
        ipywidgets.IntSlider
        """
        # Get number of all force extension pairs
        stop = len(self.stress_release_pairs(**kwargs)[0])

        def view_image(i):
            try:
                self.plot_force_extension(i, **kwargs)
            except:
                print('No image found!')

        slider = IntSlider(min=0, max=stop - 1, step=1, value=0,
                           description='Image:')

        return interact(view_image, i=slider)

    def save_force_extension_plots(self, directory=None, file_prefix=None,
                                   file_suffix=None, file_extension='.png',
                                   **kwargs):
        """
        Save all plots created by `plot_force_extensions()`.

        directory : str
            The directory the images to be displayed are located in.
        file_prefix : str
            Display only the files beginning with `prefix`.
        file_suffix : str
            Display only the files ending with `suffix`.
        file_extension : str, optional
            The extension of the images that should be displayed. Default is
            '.png'.
        figure : matplotlib figure, optional
            A reference to a figure that should be used to plot the force
            extension pairs. If no figure is given, a new one is automatically
            created.
        **kwargs
            Parameters passed to the method `self.plot_force_extensions()`.
        """
        kwargs.pop('draw', None)
        # Create generator for all force/extension stress/release pairs
        figures = self.plot_force_extensions(draw=False, **kwargs)

        # Save all figures
        evaluate.save_figures(figures, directory=directory,
                              file_prefix=file_prefix, file_suffix=file_suffix,
                              file_extension=file_extension)

        # Redraw the figure, after the last one has been saved
        self.fe_figure.canvas.draw()

    @property
    def displacementXYZ(self):
        """
        Displacement in µm with height dependent calibration factors for X, Y
        and Z.
        """
        data = self.get_data(traces=['psdXYZ', 'positionZ'])
        psdXYZ = data[:, 0:3]
        positionZ = data[:, [3]]
        displacementXYZ = self.calibration.displacement(psdXYZ,
                                                        positionZ=positionZ)
        return displacementXYZ

    @property
    def forceXYZ(self):
        """
        Force in nN, that is acting on the tether
        """
        positionZ = self.get_data(traces='positionZ')
        forceXY_Z = self.calibration.force(self.displacementXYZ,
                                           positionZ=positionZ)
        return forceXYZ(forceXY_Z)

    def _force(self, samples=None, twoD=False):
        """
        Magnitude of the force in nN acting on the tethered molecule (1D
        numpy.ndarray).
        """
        # Get force (in a fast way)
        data = self.get_data(traces=['psdXYZ', 'positionXYZ'], samples=samples)

        psdXYZ = data[:, 0:3]
        positionXY = data[:, 3:5]
        positionZ = data[:, [5]]

        displacementXYZ \
            = self.calibration.displacement(psdXYZ, positionZ=positionZ)
        # 2D or 3D calculation of the distance in Z
        if twoD:
            displacementXYZ[:, Z] = 0.0
        forceXY_Z = self.calibration.force(displacementXYZ,
                                           positionZ=positionZ)

        fXYZ = forceXYZ(forceXY_Z)
        f = force(fXYZ, positionXY)

        return f

    @property
    def force(self):
        """
        Magnitude of the force in nN determined in 3D, acting on the tethered
        molecule.
        """
        return self._force()

    @property
    def distanceXYZ(self):
        """
        Distance of the attachment point to the bead center for all 3 axes.
        """
        # µm, point of attachment of DNA
        positionXYZ = self.get_data(traces='positionXYZ')
        return distanceXYZ(positionXYZ, self.displacementXYZ,
                           self.calibration.radius,
                           self.calibration.focalshift)

    @property
    def distance(self):
        """
        Distance of the attachment point to the bead center.
        """
        positionXY = self.get_data(traces='positionXY')
        return distance(self.distanceXYZ, positionXY)

    def _extension(self, samples=None, twoD=False):
        """
        Extension in µm of the tethered molecule (1D numpy.ndarray).
        """
        # Get extension (in a fast way)
        data = self.get_data(traces=['psdXYZ', 'positionXYZ'], samples=samples)

        psdXYZ = data[:, 0:3]
        positionXYZ = data[:, 3:6]
        positionXY = positionXYZ[:, 0:2]
        positionZ = data[:, [5]]

        displacementXYZ \
            = self.calibration.displacement(psdXYZ, positionZ=positionZ)
        # 2D or 3D calculation of the distance in Z
        if twoD:
            displacementXYZ[:, Z] = 0.0
        distXYZ = distanceXYZ(positionXYZ, displacementXYZ,
                              self.calibration.radius,
                              self.calibration.focalshift)

        dist = distance(distXYZ, positionXY)
        e = extension(dist, self.calibration.radius)

        return e

    @property
    def extension(self):
        """
        Extension of the tethered molecule in µm.
        """
        return self._extension()

    def _force_extension(self, samples=None, twoD=False):
        """
        Extension (µm, first column) of and force (nN, second column) acting
        on the tethered molecule (2D numpy.ndarray).
        """
        # Get extension and force (in a fast way)
        data = self.get_data(traces=['psdXYZ', 'positionXYZ'], samples=samples)

        psdXYZ = data[:, 0:3]
        positionXYZ = data[:, 3:6]
        positionXY = positionXYZ[:, 0:2]
        positionZ = data[:, [5]]

        displacementXYZ \
            = self.calibration.displacement(psdXYZ, positionZ=positionZ)
        # 2D or 3D calculation of the distance in Z
        if twoD:
            displacementXYZ[:, Z] = 0.0
        distXYZ = distanceXYZ(positionXYZ, displacementXYZ,
                              self.calibration.radius,
                              self.calibration.focalshift)

        dist = distance(distXYZ, positionXY)
        forceXY_Z = self.calibration.force(displacementXYZ,
                                           positionZ=positionZ)

        fXYZ = forceXYZ(forceXY_Z)

        e = extension(dist, self.calibration.radius)
        f = force(fXYZ, positionXY)

        return np.c_[e, f]

    @property
    def force_extension(self):
        """
        Extension (µm, first column) of and force (nN, second column) acting
        on the tethered molecule (2D numpy.ndarray).
        """
        return self._force_extension()

    @property
    def angle(self):
        """
        Returns a dictionary of angles, for three corners (A), (B), and (C),
        calculated by (F)orce and (D)istance:
            B
            |\
          a | \ c
            |  \
            |___\
           C  b  A
        """
        return angle(self.force, self.forceXYZ, self.excited_axis,
                     self.distance, self.distanceXYZ)

    @property
    def rightstress(self):
        return self.sections(direction='right', cycle='stress',
                             range_concat=True)

    @property
    def leftstress(self):
        return self.sections(direction='left', cycle='stress',
                             range_concat=True)

    @property
    def leftrelease(self):
        return self.sections(direction='left', cycle='release',
                             range_concat=True)

    @property
    def rightrelease(self):
        return self.sections(direction='right', cycle='release',
                             range_concat=True)

    @property
    def stress(self):
        return self.sections(cycle='stress', range_concat=True)

    @property
    def release(self):
        return self.sections(cycle='release', range_concat=True)

    @property
    def rightstresses(self):
        return self.sections(direction='right', cycle='stress')

    @property
    def leftstresses(self):
        return self.sections(direction='left', cycle='stress')

    @property
    def leftreleases(self):
        return self.sections(direction='left', cycle='release')

    @property
    def rightreleases(self):
        return self.sections(direction='right', cycle='release')

    @property
    def stresses(self):
        return self.sections(cycle='stress')

    @property
    def releases(self):
        return self.sections(cycle='release')


# Define constants for convenient handling
X = 0
Y = 1
Z = 2
XY = hp.slicify([X, Y])
XZ = hp.slicify([X, Z])
YZ = hp.slicify([Y, Z])
XYZ = hp.slicify([X, Y, Z])


def forceXYZ(forceXY_Z, copy=True):
    """
    Force in nN, that is acting on the tether
    """
    if copy:
        forceXYZ = forceXY_Z.copy()
    else:
        forceXYZ = forceXY_Z
    # stressing on the bead (negative displacement) corresponds to a positive
    # force in Z
    forceXYZ[:, Z] = - 1.0 * forceXY_Z[:, Z]
    return forceXYZ


def force(forceXYZ, positionXY):
    """
    Returns force.

    Parameters
    ----------
    forceXYZ : 2D numpy.ndarray of type float
        forceXYZ.shape[1] can consist of either 3 (XYZ) or 2 (XY) axes
    """
    # sign of forceXY depends on positionXY, important for noise of force
    # around +/- 0
    # forceZ negative/positive, irrespective of positionZ!
    signF = np.sign(forceXYZ)
    signF[:, XY] = np.sign(forceXYZ[:, XY]) * np.sign(positionXY)

    # square the forces and account for the signs
    force_sq = forceXYZ**2 * signF
    forceSUM = np.sum(force_sq, axis=1)
    force = np.sqrt(np.abs(forceSUM)) * np.sign(forceSUM)
    return force


def distanceXYZ(positionXYZ, displacementXYZ, radius=0.0, focalshift=1.0,
                clip_Z=True):
    """
    Distance of the attachment point to the bead center for all 3 axes.
    positionXYZ, displacementXYZ and radius need to have the same unit.
    If radius is 0.0, the positionZ is not corrected by the radius. You should
    set radius to 0.0, if the positionZ is defined in such a way, that
    positionZ is 0.0, where the bead center would be on the glass surface.
    If focalshift is 1.0, the positionXYZ is not corrected by the focalshift.

    Parameters
    ----------
    positionXYZ : 2D np.array of type float
    displacementXYZ : 2D np.array of type float
    radius : float
    focalshift : float
    clip_Z : bool
        The distance of the attachment point to the center of the bead cannot
        be smaller than the radius. Therefore, clip the data to be at least as
        great as the radius.
        However, values much smaller than the radius could indicate an errornes
        calibration with too small displacement sensitivities, which would lead
        to too small displacements in Z and in turn to negative distances.
        Therefore, if you want to check for this kind of error, switch off the
        clip_Z functionality.
    """
    # distance, point of attachment of DNA
    # displacement, displacement of bead out of trap center
    # radius
    distanceXYZ = positionXYZ.copy()
    # distance from attachment point to center of bead
    # attachmentXY - displacementXY
    distanceXYZ[:, 0:2] -= displacementXYZ[:, 0:2]

    # If the bead is free (i.e. above the surface with a distance Z > 0), a
    # movement of the positionZ leads to a distance change reduced by the focal
    # shift.
    # If the bead is on the surface, a movement of the positionZ leads to a
    # distance change independant of the focal shift.
    idx_free = positionXYZ[:, 2] < 0
    idx_touch = positionXYZ[:, 2] >= 0
    distanceXYZ[idx_free, 2] = (- positionXYZ[idx_free, 2] * focalshift
                                + radius
                                # distanceZ + radius + displacementZ
                                + displacementXYZ[idx_free, 2])
    distanceXYZ[idx_touch, 2] = (- positionXYZ[idx_touch, 2]
                                 + radius
                                 # distanceZ + radius + displacementZ
                                 + displacementXYZ[idx_touch, 2])

    if clip_Z:
        distanceXYZ[:, 2] = distanceXYZ[:, 2].clip(min=radius)
    # A positive positionZ signal (positionZ upwards) corresponds to a
    # decreasing (negative) distance of the bead to the surface:
    #   -> distanceZ ~ - positionZ
    # A movement of the positionZ gets reduced by the focalshift:
    #   -> distanceZ = - positionZ * focalshift
    # The distanceZ (positionZ) is 0, where the bead touches the surface:
    #   -> center of the bead is at distanceZ + radius
    # The bead is stressed down with increasing positive distanceZ. This leads
    # to a negative displacement, which reduces the distance:
    #   -> distanceZ + displacement

    # distance from attachment point to bead center
    return distanceXYZ


def distance(distanceXYZ, positionXY):
    """
    Calculate the distance of the attachment point to the bead center.

    Parameters
    ----------
    distanceXYZ : 2D numpy.ndarray of type float
        distanceXYZ.shape[1] can consist of either 3 (XYZ) or 2 (XY) axes
    """
    # sign of distanceXY depends on positionXY, important for noise of dist
    # around +/- 0
    # distanceZ negative/positive, irrespective of positionZ!
    signD = np.sign(distanceXYZ)
    signD[:, XY] = np.sign(distanceXYZ[:, XY]) * np.sign(positionXY)

    # square the distances and account for the signs
    distance_sq = distanceXYZ**2 * signD
    distSUM = np.sum(distance_sq, axis=1)
    return np.sqrt(np.abs(distSUM)) * np.sign(distSUM)


def extension(distance, radius):
    """
    Calculate the extension of the DNA by simply subtracting the radius from
    the distance.
    """
    return distance - radius


def angle(force, forceXYZ, excited_axis, distance, distanceXYZ):
    """
    Returns a dictionary of angles, for three corners (A), (B), and (C),
    calculated by (F)orce and (D)istance:
        B
        |\
      a | \ c
        |  \
        |___\
       C  b  A
    """

    Fabs = force
    Fz = forceXYZ[:, 2]
    if excited_axis == 'X':
        ea = 0
    else:
        ea = 1
    Fxy = forceXYZ[:, ea]

    dist = distance
    dist3D = distanceXYZ
    Dz = dist3D[:, 2]
    Dxy = dist3D[:, ea]

    AF = _angle(Fz, Fxy, Fabs, angle='A')
    BF = _angle(Fz, Fxy, Fabs, angle='B')
    CF = _angle(Fz, Fxy, Fabs, angle='C')
    AD = _angle(Dz, Dxy, dist, angle='A')
    BD = _angle(Dz, Dxy, dist, angle='B')
    CD = _angle(Dz, Dxy, dist, angle='C')

    angle = dict(list(zip(['AF', 'BF', 'CF', 'AD', 'BD', 'CD'],
                          [AF, BF, CF, AD, BD, CD])))

    return angle


def _angle(a, b, c, angle='A'):
    """
        B
        |\
      a | \ c
        |  \
        |___\
       C  b  A
    """
    if angle == 'A':
        _a = b
        _b = c
        _c = a
    elif angle == 'B':
        _a = a
        _b = c
        _c = b
    else:
        _a = a
        _b = b
        _c = c

    cos_angle = (_a**2 + _b**2 - _c**2) / (2 * _a * _b)
    return np.arccos(np.fabs(cos_angle)) * 180.0 / np.pi


def _force_extension_figure(grid=True):
    """
    Create a figure to plot force/extension stress/release pair(s).

    Parameters
    ----------
    grid : bool
        Switch on/of the grid of the axis.

    Returns
    -------
    matplotlib.figure.Figure
        Return a reference to the figure.
    """
    # Create figure and axis for plotting
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    # Set labels
    ax.set_xlabel("Extension (nm)")
    ax.set_ylabel("Force (pN)")

    # Switch on/off grid
    ax.grid(grid)

    # Create lines for stress/release extension/force data pairs
    # line_stress =
    ax.plot([0], [0], 'm.', ms=1.0)
    # line_release =
    ax.plot([0], [0], 'c.', ms=1.0)
    # standard curve for DNA
    ax.plot([0], [0], 'r')

    # Plot stress/release extension/force data pairs
    title = "Force Extension"
    ax.set_title(title)

    return fig
